local XPTracker = LibStub("AceAddon-3.0"):GetAddon("XPTracker")
local L = LibStub("AceLocale-3.0"):GetLocale("XPTracker")

local ConfigMenu = XPTracker:GetModule("ConfigMenu")
local Widgets = XPTracker:GetModule("Widgets")

local function getOptions()
  local profile = XPTracker.db.profile
  options = {
    name = "XPTracker",
    type = "group",
    args = {
      general = {
        type = "group",
        inline = true,
        name = "",
        args = {
          lockWindow = {
            name = "Lock window",
            desc = "Lock the GUI window in place",
            type = "toggle",
            order = 1,
            set = function(info, val)
              profile.MainWindow.IsLocked = val
              ConfigMenu:HandleMainWindowMovement()
            end,
            get = function(info) return profile.MainWindow.IsLocked end
          },
          hideWindow = {
            name = "Toggle visibility",
            desc = "Toggle hiding the GUI window",
            type = "toggle",
            order = 2,
            set = function(info, val)
              profile.MainWindow.ShowingWindow = val
              ConfigMenu:ToggleWindowVisibility()
            end,
            get = function(info, val) return profile.MainWindow.ShowingWindow end
          },
          moveToCenter = {
            name = "Center GUI window",
            desc = "Move window to center in case you lost track of it",
            type = "execute",
            func = function()
              ConfigMenu:BringWindowToCenter()
            end,
            order = 3,
          }
        }
      }
    },
  }
  return options
end

function ConfigMenu:HandleMainWindowMovement()
  local isLocked = XPTracker.db.profile.MainWindow.IsLocked
  XPTracker.MainWindow:SetMovable(not isLocked)
  if isLocked then
    XPTracker.MainWindow:RegisterForDrag()
  else
    XPTracker.MainWindow:RegisterForDrag("LeftButton")
  end
end

function ConfigMenu:BringWindowToCenter()
  local windowPosition = XPTracker.db.profile.MainWindow.Position
  local window = XPTracker.MainWindow
  local halfScreenWidth = window:GetParent():GetWidth() / 2
  local halfScreenHeight = window:GetParent():GetHeight() / 2
  window:SetPoint("TOPLEFT", window:GetParent(), "TOPLEFT", halfScreenWidth, -halfScreenHeight);
  Widgets:UpdateWindowPosition(window)
end

function ConfigMenu:ToggleWindowVisibility()
  local showingWindow = XPTracker.db.profile.MainWindow.ShowingWindow
  local window = XPTracker.MainWindow
  if showingWindow then
    window:Show()
  else
    window:Hide()
  end
end

function ConfigMenu:RegisterConfigMenu()
  LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("XPTracker Options", getOptions())
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions("XPTracker Options", "XPTracker", nil, "general")
end
